# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
from datetime import datetime
import cellregulondb

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "cellregulondb"
copyright = f"{datetime.now():%Y}, Teichmann Lab"
author = "Jan Patrick Prett <jp30@sanger.ac.uk>"
release = cellregulondb.__version__

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath(".."))

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
]
source_suffix = [".rst", ".md"]
master_doc = "index"
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

autosummary_generate = True
napoleon_google_docstring = True
napoleon_use_rtype = True
napoleon_use_param = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
#html_static_path = ["_static"]

intersphinx_mapping = dict(
    python=("https://docs.python.org/3/", None),
    numpy=("https://numpy.org/doc/stable/", None),
    pandas=("https://pandas.pydata.org/pandas-docs/stable/", None),
)



#!/usr/bin/env python3

import warnings

warnings.filterwarnings("ignore")

import json
import sqlite3
import logging
import requests
import pandas as pd

logging.basicConfig(level="INFO", format="[%(asctime)s][%(levelname)s] %(message)s")


###########################################################
# Config section
###########################################################

# DB name
DB_PATH = f"cellregulon_2023-06-22.db"

# csv output path
CSV_PATH = f"gene-match-cellregulon.csv"

# update database or only generate output csv report
UPDATE_DATABASE = True

# fetch symbols and update db
# TODO: split this into 2 functions to make it cleaner?
def fetch_symbols():
    # open db connection
    connection = sqlite3.connect(DB_PATH)
    cursor = connection.cursor()

    # make everything fast — what? that it may lead to corruption in the DB? oh well...
    connection.execute("PRAGMA synchronous  = OFF")
    connection.execute("PRAGMA journal_mode = MEMORY")
    connection.execute("PRAGMA temp_store   = memory")
    connection.execute("PRAGMA mmap_size    = 30000000000")

    # get all symbols from the DB
    all_symbols = cursor.execute(
        "SELECT id, name, properties->>'is_TF' FROM nodes"
    ).fetchall()
    match_results = []

    if not UPDATE_DATABASE:
        cursor.close()
        connection.close()

    # setup both GRCh38 and GRCh37 endpoints and headers
    GRCH38_ENSEMBL_API = "https://rest.ensembl.org/lookup/symbol/homo_sapiens"
    GRCH37_ENSEMBL_API = "https://grch37.rest.ensembl.org/lookup/symbol/homo_sapiens"
    HEADERS = {"Content-Type": "application/json", "Accept": "application/json"}

    # first pass will try to match symbol names to GRCh38
    total = len(all_symbols)
    good = 0
    bad = 0
    failed = []
    # batch genes 1000 at the time
    for i in range(0, len(all_symbols), 999):
        batch_symbols = all_symbols[i : i + 999]
        response = requests.post(
            GRCH38_ENSEMBL_API,
            headers=HEADERS,
            json={"symbols": [s[1] for s in batch_symbols]},
        )
        if response.ok:
            data = response.json()
            for symbol in batch_symbols:
                if symbol[1] in data.keys():
                    if UPDATE_DATABASE:
                        node = cursor.execute(
                            "SELECT id, properties FROM nodes WHERE id = ?", [symbol[0]]
                        ).fetchone()
                        props = json.loads(node[1])
                        props["symbol_information"] = data[symbol[1]]
                        cursor.execute(
                            "UPDATE nodes SET properties = ? WHERE id = ?",
                            [json.dumps(props), node[0]],
                        )
                    good += 1
                    match_results.append(
                        {"gene_in_db": symbol[1], "is_TF": symbol[2], **data[symbol[1]]}
                    )
                else:
                    bad += 1
                    failed.append(symbol)
            logging.info(f"[Grch38] Match={good}. Unmatched={bad}. Total={total}.")
        else:
            logging.error("[Grch38] POST failed", response.content)
        if UPDATE_DATABASE:
            connection.commit()
    logging.info(f"[Grch38] Finished matching symbols")

    # second pass will try to match unmatched symbol names to GRCh37
    logging.info(f"Running failed symbols {len(failed)} with Grch37.")
    total = len(failed)
    good = 0
    bad = 0
    for i in range(0, len(failed), 999):
        batch_symbols = failed[i : i + 999]
        response = requests.post(
            GRCH37_ENSEMBL_API,
            headers=HEADERS,
            json={"symbols": [s[1] for s in batch_symbols]},
        )
        if response.ok:
            data = response.json()
            for symbol in batch_symbols:
                if symbol[1] in data.keys():
                    if UPDATE_DATABASE:
                        node = cursor.execute(
                            "SELECT id, properties FROM nodes WHERE id = ?", [symbol[0]]
                        ).fetchone()
                        props = json.loads(node[1])
                        props["symbol_information"] = data[symbol[1]]
                        cursor.execute(
                            "UPDATE nodes SET properties = ? WHERE id = ?",
                            [json.dumps(props), node[0]],
                        )
                    good += 1
                    match_results.append(
                        {"gene_in_db": symbol[1], "is_TF": symbol[2], **data[symbol[1]]}
                    )
                else:
                    bad += 1
                    match_results.append({"gene_in_db": symbol[1], "is_TF": symbol[2]})
            logging.info(f"[GRCh37] Match={good}. Unmatched={bad}. Total={total}.")
        else:
            logging.error("[GRCh37] POST fail", response.content)

        if UPDATE_DATABASE:
            connection.commit()

    # gracefuly close everything
    if UPDATE_DATABASE:
        cursor.close()
        connection.close()

    # return dataframe with all match results
    return pd.DataFrame.from_records(match_results)


if __name__ == "__main__":
    result = fetch_symbols()
    logging.info(f"Finished looking up symbols")

    df = pd.DataFrame.from_records(result)
    total = df.shape[0]
    matched = df[pd.isnull(df["id"])].shape[0]
    unmatched = df[~pd.isnull(df["id"])].shape[0]
    tf_total = df[df["is_TF"] == 1].shape[0]
    tf_matched = df[~pd.isnull(df["id"]) & df["is_TF"] == 1].shape[0]
    tf_unmatched = df[pd.isnull(df["id"]) & df["is_TF"] == 1].shape[0]

    logging.info(f"[TF] matched={tf_matched}")
    logging.info(f"[TF] unmatched={tf_unmatched}")
    logging.info(f"[TF] total={tf_total}")

    logging.info(f"[all] matched={matched} ({round(matched*100/total,2)}%)")
    logging.info(f"[all] unmatched={unmatched} ({round(unmatched*100/total,2)}%)")
    logging.info(f"[all] total={total}")

    logging.info(f"Saving results as '{CSV_PATH}'")
    df.to_csv(CSV_PATH, index=False)

    logging.info(f"Done")



from typing import Union
import os
import json
import pickle
import logging
import sqlite3
import pandas as pd
import networkx as nx

logging.basicConfig(level="INFO", format="[%(asctime)s][%(levelname)s] %(message)s")


class CellRegulonDB:
    """
    A class wrapping the CellRegulon database.

    Args:
        db_path (str): Path to the SQLite database file.

    Attributes:
        db (sqlite3.Connection): Connection to the SQLite database.
        TF (pandas.DataFrame): DataFrame containing information about transcription factors (TFs).
    """

    def __init__(self, db_path: str):
        """
        Initializes a CellRegulonDB object.

        Args:
            db_path (str): Path to the SQLite database file.

        Raises:
            Exception: If the provided database file does not exist.

        """
        if not os.path.isfile(db_path):
            raise Exception(f"Invalid file '{db_path}'")
        self.db = sqlite3.connect(db_path)
        self.TFs = self._build_TF_dataframe()
        self.lineages = self._get_lineages()

    def _get_lineages(self) -> pd.DataFrame:
        rows = self.db.execute(
            "select distinct properties->>'cell_label' from edges"
        ).fetchall()
        return pd.DataFrame([row[0] for row in rows], columns=["lineage"])

    def _build_TF_dataframe(self) -> pd.DataFrame:
        """
        Builds a DataFrame containing information about transcription factors (TFs) from the database.

        Returns:
            pandas.DataFrame: DataFrame containing TF information.

        """
        rows = self.db.execute(
            """
            select name, properties 
            from nodes where properties->>'is_TF'=1
            """
        ).fetchall()

        return pd.DataFrame.from_records(
            [
                self._get_symbol_info(row[0], json.loads(row[1])).values()
                for row in rows
            ],
            columns=["TF", "ENS", "chromosome", "start", "end"],
        )

    def location_search(
        self,
        chromosome: str,
        location: int = None,
        start: int = None,
        end: int = None,
        delta: int = 10000,
    ) -> pd.DataFrame:
        """
        Searches for transcription factors based on location information.

        Args:
            chromosome (str): Chromosome name.
            location (int, optional): Specific location to search for the start of the transcription factors. Defaults to None.
            start (int, optional): Start position of the range to search for transcription factors. Defaults to None.
            end (int, optional): End position of the range to search for transcription factors. Defaults to None.
            delta (int, optional): Delta value used for range search. Defaults to 10000 (+/-10kb).

        Returns:
            pandas.DataFrame: DataFrame containing the matching transcription factors.

        Raises:
            Exception: If neither 'location' nor 'start' and 'end' are provided.

        """
        chromosome = str(chromosome)

        if location is not None:
            return self.TFs[
                (location >= self.TFs["start"] - delta)
                & (location <= self.TFs["start"] + delta)
                & (self.TFs["chromosome"] == chromosome)
            ]
        elif start is not None and end is not None:
            return self.TFs[
                (end >= self.TFs["start"] - delta)
                & (start <= self.TFs["start"] + delta)
                & (self.TFs["chromosome"] == chromosome)
            ]
        else:
            raise Exception("Required 'location' or 'start' and 'end'")

    def _get_regulon(self, TF: str, lineage: list = None) -> pd.DataFrame:
        """
        Retrieves the regulon for a given transcription factor.

        Args:
            TF (str): Name of the transcription factors.
            lineage (list, optional): List of cell lineages to filter the regulon. Defaults to None.

        Returns:
            pandas.DataFrame: DataFrame containing the regulon information.

        """
        regulon = []
        rows = self.db.execute(
            """
            select s.name,t.name,e.properties
            from edges as e, nodes as s, nodes as t
            where s.id=e.source_id and t.id=e.target_id and s.name = ?
            """,
            [TF],
        ).fetchall()
        for row in rows:
            data = json.loads(row[2])
            if lineage is None or data["cell_label"] in lineage:
                regulon.append(
                    [
                        row[0],
                        row[1],
                        data["regulation"],
                        data["coexpresion"],
                        data["motif_enrichment"],
                        data["cell_label"],
                        data["dataset"],
                    ]
                )
        return pd.DataFrame(
            regulon,
            columns=[
                "TF",
                "gene",
                "regulation",
                "coexpresion",
                "motif_enrichment",
                "cell_label",
                "dataset_source",
            ],
        )

    def get_regulon(self, TF: Union[str, list], lineage: list = None) -> pd.DataFrame:
        """
        Retrieves the regulon for one or more TFs.

        Args:
            TF (Union[str, list]): Name of the TF or list of TF names.
            lineage (list, optional): List of cell lineages to filter the regulon. Defaults to None.

        Returns:
            pandas.DataFrame: DataFrame containing the regulon information.

        """
        if isinstance(TF, list):
            regulons = pd.concat([self._get_regulon(tf, lineage) for tf in set(TF)])
        else:
            regulons = self._get_regulon(str(TF), lineage)
        return regulons

    def get_shared_genes(self, data: Union[list, pd.DataFrame]) -> list:
        """
        Retrieves the shared genes among multiple regulons.

        Args:
            data (Union[list, pd.DataFrame]): List of transcription factors or a DataFrame containing regulon information.

        Returns:
            list: A list containing the shared genes.

        """
        if isinstance(data, list):
            data = self.get_regulon(data)
        regulons = data.groupby("TF")["gene"].apply(list)
        return list(set.intersection(*[set(genes) for _, genes in regulons.items()]))

    def _get_symbol_info(self, symbol: str, attr: dict) -> dict:
        """
        Extracts relevant information from a node's attributes.

        Args:
            symbol (str): Name of the gene.
            attr (dict): Dictionary containing symbol attributes.

        Returns:
            dict: Dictionary with extracted symbol information.

        """
        info = {"TF": symbol, "ENS": 0, "chromosome": 0, "start": 0, "end": 0}
        if "symbol_information" in attr and attr["symbol_information"]:
            _id = attr["symbol_information"]["id"]
            _start = attr["symbol_information"]["start"]
            _end = attr["symbol_information"]["end"]
            _strand = attr["symbol_information"]["strand"]
            _chromosome = attr["symbol_information"]["seq_region_name"]
            info["ENS"] = _id
            info["chromosome"] = _chromosome
            info["start"] = _start if _strand == 1 else _end
            info["end"] = _end if _strand == 1 else _start

        return info

    def to_networkx(
        self, df: pd.DataFrame = None, source: str = "TF", target: str = "gene"
    ) -> nx.MultiDiGraph:
        """
        Converts the database to a NetworkX MultiDiGraph.

        Returns:
            nx.MultiDiGraph: NetworkX MultiDiGraph representing the database.

        """
        if df is None:
            logging.warning(
                "No df provided. Converting whole database to NetworkX MultiDiGraph. This operation will take a long time."
            )
        # create graph
        G = nx.MultiGraph()
        if df is None:
            # read all nodes
            nodes = self.db.execute("select id, name, properties from nodes").fetchall()
            # add them to the graph
            G.add_nodes_from([(n[1], json.loads(n[2])) for n in nodes])

            # read all edges
            edges = self.db.execute(
                """
                select s.name,t.name,e.properties
                from edges e, nodes t, nodes s
                where s.id=e.source_id and t.id=e.target_id
                """
            ).fetchall()
            # add them to the graph
            G.add_edges_from([(e[0], e[1], json.loads(e[2])) for e in edges])
        else:
            G.add_edges_from(
                [
                    (data[source], data[target], data.to_json(orient="columns"))
                    for _, data in df.iterrows()
                ]
            )

        return G

    def save_networkx(G, path: str = "cellregulon.nx") -> None:
        # store as pickled object
        with open(path, "wb") as f:
            pickle.dump(G, f)

    def load_networkx(path: str = "cellregulon.nx") -> nx.MultiDiGraph:
        # load pickled object
        with open(path, "rb") as f:
            G = pickle.load(f)
        return G

    def to_cytosscape(self, df: pd.DataFrame, name: str = "cytoscape.txt") -> None:
        """
        Exports the database to a Cytoscape-readable file.

        Args:
            name (str, optional): Name of the output file. Defaults to "cytoscape.txt".

        """
        # import networkx
        # self.G.cytoscape_data
        pass

    def as_pyscenic(self, df: pd.DataFrame) -> "pySCENIC":
        try:
            import pyscenic
        except ModuleNotFoundError:
            raise Exception("Module 'pyscenic' is not installed.")
        """
        Converts the database to pySCENIC format.

        Returns:
            pySCENIC?: pySCENIC representation of the database.

        """
        pass


def download_db(db_path: str = None, version: str = "latest") -> str:
    """
    Downloads a CellRegulon database file.

    Args:
        db_path (str, optional): Path to save the downloaded database file. If not provided, the file will be saved as "cellregulon-vXXX.db" in the current directory. Defaults to None.
        version (str, optional): Version of the CellRegulon database to download. Defaults to "latest".

    Returns:
        str: Path to the downloaded database file.

    Raises:
        Exception: If the specified version is not available.

    """
    import requests
    from rich.progress import (
        BarColumn,
        DownloadColumn,
        Progress,
        TimeRemainingColumn,
        TransferSpeedColumn,
    )

    progress_columns = (
        "[progress.description]{task.description}",
        BarColumn(),
        TimeRemainingColumn(),
        DownloadColumn(),
        TransferSpeedColumn(),
    )

    # get avaiable versions
    versions_url = "https://cellregulondb.cog.sanger.ac.uk/db/versions.json"
    data = requests.get(versions_url).json()
    if version == "latest":
        version = data["latest"]
    required_version = [db for db in data["versions"] if version == db["version"]]

    # sanity check the version exists
    if not required_version:
        raise Exception(
            f"Wrong verision number. Avaiable versions: {[db['version'] for db in data['versions']]}"
        )

    selected = required_version[0]
    if not db_path:
        db_path = f"cellregulon-v{selected['version']}.db"

    if os.path.isfile(db_path):
        logging.warning(
            f"Database file '{db_path}' already exists. Will be overwritten."
        )

    with requests.get(selected["url"], allow_redirects=True, stream=True) as r:
        r.raise_for_status()
        content_length = int(r.headers.get("Content-Length"))
        logging.info(f"Downloading database verssion {selected['version']}")
        with Progress(*progress_columns) as progress:
            task = progress.add_task(
                f"downloading v{selected['version']}", total=content_length
            )
            with open(db_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=5 * 1024**2):
                    progress.advance(task, f.write(chunk))
            logging.info(f"Downloaded {db_path}")
            return db_path



#!/usr/bin/env python3

import warnings

warnings.filterwarnings("ignore")
import os
import json
import time
import sqlite3
import logging
import datetime
import loompy as lp
import pandas as pd
from pyscenic.cli.utils import load_signatures

logging.basicConfig(level="INFO", format="[%(asctime)s][%(levelname)s] %(message)s")


###########################################################
# Config section
###########################################################

# path to the results manifests
# expected manifest fromat
# | dataset | dataset_lineage | ontology_label |      AUCell      |    GRNBoost   |   cisTarget   |
# |   name  |     lineage     |   onto label   |  path/file.loom  |  path/adj.csv | path/reg.csv  |
RESULTS_MANIFEST = "results.csv"

# DB name (uses 'cellregulon_<current date>.db' by default)
DB_PATH = f"cellregulon_{datetime.date.today()}.db"

# get coexpresion from GRNboost?
FETCH_COEXPRESION = True

# get motif expression from cisTarget?
FETHC_MOTIF_ENRICHMENT = True


###########################################################
# DB schema definition
###########################################################

# querys to create database structure
# - nodes table
create_nodes_table = """
CREATE TABLE IF NOT EXISTS nodes (
        id INTEGER PRIMARY KEY,
        name TEXT,
        properties JSON
);
"""
# - edges table
create_edges_table = """
CREATE TABLE IF NOT EXISTS edges (
        id INTEGER PRIMARY KEY, 
        source_id INTEGER, 
        target_id INTEGER, 
        properties JSON,
        FOREIGN KEY(source_id) REFERENCES nodes (id), 
        FOREIGN KEY(target_id) REFERENCES nodes (id)
);
"""
# - vestion table
create_version_table = """
CREATE TABLE IF NOT EXISTS version (
        version TEXT,
        notes TEXT,
        timestamp TIMESTAMP DEFAULT CURRENT_TIMESTAMP NOT NULL
);
"""


###########################################################
# Init DB and create tables and indices
###########################################################
def create_db():
    # connect to db file and create DB and indices
    logging.info(f"Opening database '{DB_PATH}'")
    connection = sqlite3.connect(DB_PATH)
    cursor = connection.cursor()
    logging.info(f"Creating database schema '{DB_PATH}'")
    cursor.execute(create_nodes_table)
    cursor.execute("CREATE INDEX IF NOT EXISTS idx_nodes ON nodes (id);")
    cursor.execute("CREATE INDEX IF NOT EXISTS idx_nodes_name ON nodes (name);")
    cursor.execute(create_edges_table)
    cursor.execute("CREATE INDEX IF NOT EXISTS idx_edges ON edges (id);")
    cursor.execute("CREATE INDEX IF NOT EXISTS idx_edges_source ON edges (source_id);")
    cursor.execute("CREATE INDEX IF NOT EXISTS idx_edges_target ON edges (target_id);")
    cursor.execute(create_version_table)
    logging.info(f"Done creating database '{DB_PATH}'")
    connection.commit()
    cursor.close()
    connection.close()


###########################################################
# Manifest processing and inserts
###########################################################
def main():
    connection = sqlite3.connect(DB_PATH)
    cursor = connection.cursor()

    # make everything fast — what? that it may lead to corruption in the DB? oh well...
    connection.execute("PRAGMA synchronous  = OFF")
    connection.execute("PRAGMA journal_mode = MEMORY")
    connection.execute("PRAGMA temp_store   = memory")
    connection.execute("PRAGMA mmap_size    = 30000000000")

    # keep a references of all the inserted nodes
    nodes = {}

    # process manifest file
    df = pd.read_csv(RESULTS_MANIFEST, index_col=False)
    time_process = time.time()
    for _, row in df.iterrows():
        time_row = time.time()
        dataset = row["dataset"]
        lineage = row["dataset_lineage"]
        DOI = row["DOI"]
        CL = row["ontology_label"]
        logging.info(f"Dataset: '{dataset}'. Lineage '{lineage}'. Ontology '{CL}'")

        # read loom output from AUCell
        auc_outpout = row["AUCell"]
        logging.info(f"Reading AUCell output: '{auc_outpout}'")
        lf = lp.connect(auc_outpout, mode="r", validate=False)

        # create dataframe
        # some examples at https://github.com/aertslab/pySCENIC/issues/352
        df = pd.DataFrame(lf.ra.Regulons, index=lf.ra.Gene)
        logging.info(f"Total Regulons in file: {df.shape[1]}")
        logging.info(f"Total Genes in file: {df.shape[0]}")

        # remove genes that don't match any regulon — why are those here tho?
        df = df.loc[~(df == 0).all(axis=1)]
        logging.info(f"Genes after filtering: {df.shape[0]}")

        # insert all nodes first
        logging.info(f"Inserting nodes (TF and genes)")
        # - all transcription factors
        for TF in sorted(
            list(set([k[:-3] for k in lf.ra.Regulons.dtype.fields.keys()]))
        ):
            if TF not in nodes:
                props = {"is_TF": True}
                cursor.execute(
                    "INSERT INTO nodes (name, properties) VALUES (?,?)",
                    [TF, json.dumps(props)],
                )
                nodes[TF] = cursor.lastrowid
            else:
                # if symbol was already inserted but as 'gene' instead of TF update record to be TF (is_TF = True)
                node = cursor.execute(
                    "SELECT id, properties FROM nodes WHERE id = ? AND properties->>'is_TF' = 0",
                    [nodes[TF]],
                ).fetchone()
                if node:
                    props = json.loads(node[1])
                    props["is_TF"] = True
                    cursor.execute(
                        "UPDATE nodes SET properties = ? WHERE id = ?",
                        [json.dumps(props), node[0]],
                    )

        # - all genes
        for gene in lf.ra.Gene:
            if gene not in nodes:
                props = {"is_TF": False}
                cursor.execute(
                    "INSERT INTO nodes (name, properties) VALUES (?,?)",
                    [gene, json.dumps(props)],
                )
                nodes[gene] = cursor.lastrowid
        logging.info(f"Done inserting nodes")

        # read additional analysis for this dataset/lineage
        if FETCH_COEXPRESION:
            GRNBoost_path = row["GRNBoost"]
            logging.info(f"Reading GRNBoost from file: '{GRNBoost_path}'")
            GRNBoost = pd.read_csv(GRNBoost_path)

        # preprocess cisTopic horrible output where motifs is an array of tuples per row in the file
        # but it's a string so it's converted to list eval(motifs) and then fattened into a single list
        if FETHC_MOTIF_ENRICHMENT:
            cisTarget_path = row["cisTarget"]
            logging.info(f"Reading cisTarget from file: '{cisTarget_path}'")
            cisTarget = load_signatures(cisTarget_path)

        logging.info(f"Inserting edges")
        # transverse the columns (tf) and the matching rows (genes) to insert edges
        for label, content in df.items():
            genes = content[content == 1].index.values
            logging.info(f"Inserting TF '{label}' ({len(genes)} genes)")
            rname = label[:-3]
            regulation = label[-2:-1]

            # filter GRNBoost and cisTarget using the TF
            if FETCH_COEXPRESION:
                TF_GRNBoost = GRNBoost.query(f"TF=='{rname}'")
            if FETHC_MOTIF_ENRICHMENT:
                TF_cisTarget = [tf for tf in cisTarget if tf.name == label]
                if len(TF_cisTarget) > 1:
                    logging.warning(
                        f"Regulon {label} is present more than once in cisTarget result?"
                    )

            for gene in genes:
                # the following two lookups (GRNboosst and cisTarget) take too long and should be optimized?
                # - get GRNboost coexpresion for gene
                coexpresion = 0
                if FETCH_COEXPRESION and not TF_GRNBoost.empty:
                    gene_GRNBoost = TF_GRNBoost.query(f"target=='{gene}'")
                    if not gene_GRNBoost.empty:
                        coexpresion = gene_GRNBoost["importance"].values[0]
                # - get cisTarget motif enrichment for gene
                motif_enrichment = 0
                if FETHC_MOTIF_ENRICHMENT and len(TF_cisTarget):
                    motif_enrichment = cisTarget[0].gene2weight.get(gene, 0)

                # build edge properties json
                props = {
                    "regulation": regulation,
                    "coexpresion": coexpresion,
                    "motif_enrichment": motif_enrichment,
                    "dataset": dataset,
                    "doi": DOI,
                    "lineage": lineage,
                    "cell_label": CL,
                }
                cursor.execute(
                    "INSERT INTO edges (source_id, target_id, properties) VALUES (?,?,?)",
                    [nodes[rname], nodes[gene], json.dumps(props)],
                )

        logging.info(f"Committing transaction")
        connection.commit()

        logging.info(
            f"Done in {time.strftime('%Hh%Mm%Ss', time.gmtime(time.time()-time_row))}"
        )

    cursor.close()
    connection.close()
    logging.info(
        f"Done in {time.strftime('%Hh%Mm%Ss', time.gmtime(time.time()-time_process))}"
    )


def is_valid_manifest():
    if not os.path.isfile(RESULTS_MANIFEST):
        logging.error(f"Missing manifest file '{RESULTS_MANIFEST}'")
        return False

    df = pd.read_csv(RESULTS_MANIFEST, index_col=False)

    logging.info(f"Validating columns")
    result = True
    for col in [
        "dataset",
        "doi",
        "dataset_lineage",
        "ontology_label",
        "AUCell",
        "GRNBoost",
        "cisTarget",
    ]:
        if col not in df.columns:
            result = False
            logging.error(f"Missing column '{col}' in manifest")
    if not result:
        return False

    result = True
    logging.info(f"Validating manifestfile references")
    for _, row in df.iterrows():
        if not os.path.isfile(row["AUCell"]):
            result = False
            logging.error(f"Missing file for AUCell {row['AUCell']}")
        if not os.path.isfile(row["GRNBoost"]):
            result = False
            logging.error(f"Missing file for GRNBoost {row['GRNBoost']}")
        if not os.path.isfile(row["cisTarget"]):
            result = False
            logging.error(f"Missing file for cisTarget {row['cisTarget']}")
    return result


if __name__ == "__main__":
    assert is_valid_manifest() == True, "Found missing input files. Abort."
    create_db()
    main()
